use std::collections::HashSet;

use futures::TryStreamExt;
use mongodb::{bson::doc, ClientSession, Collection};

// todo : api key auth

// case: export
// used an item -> demand
// mark as used
// update db

// ITEM
// VARIANTS
// QUANTIZATION
// IMPORT / EXPORT
// DEPENDENCIES
// DEMAND STATS
// SEASONAL REVIEWS

macro_rules! collect_results {
    ($res:expr) => {{
        let mut ret = vec![];

        while let Some(doc) = $res.try_next().await.unwrap() {
            ret.push(doc);
        }

        ret
    }};
}

macro_rules! get_mongo {
    () => {
        mongodb::Client::with_uri_str(std::env::var("DB_URI").unwrap())
            .await
            .unwrap()
    };
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct ItemVariantEntry {
    pub item_name: String,
    pub variant_name: String,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct ItemInventoryEntry {
    pub item_name: String,
    pub variant_name: String,
    pub origin: String,
    pub price: f64,
}

#[derive(serde::Deserialize, serde::Serialize)]
pub struct ItemEntry {
    pub name: String,
    pub category: String,
}

impl ItemEntry {
    pub fn new(doc: mdq::Document) -> Self {
        let name = std::path::Path::new(&doc.path)
            .file_stem()
            .unwrap()
            .to_str()
            .unwrap()
            .to_string();
        let category = doc
            .frontmatter
            .as_mapping()
            .unwrap()
            .get("category")
            .unwrap()
            .as_str()
            .unwrap()
            .to_string();
        Self { name, category }
    }

    pub async fn init_db(&self, mongodb: &mongodb::Client) {
        log::info!("Adding item {} to DB", self.name);
        let items: Collection<mongodb::bson::Document> =
            mongodb.database("cdb").collection("items");
        items
            .insert_one(
                mongodb::bson::doc! {
                    "_id": self.name.clone(),
                    "name": self.name.clone(),
                    "category": self.category.clone()
                },
                None,
            )
            .await
            .unwrap();
    }
}

/// Represents a single item in a collection of physical goods.
///
/// This struct serves as a blueprint for describing individual items within
/// a larger inventory or database of physical goods. It includes fields for
/// the name of the item and its category.
pub struct Item {
    db: mongodb::Client,
    pub item: ItemEntry,
}

impl Item {
    pub fn new(db: mongodb::Client, item: ItemEntry) -> Self {
        Self { db, item }
    }

    pub async fn get_inventory_entries(&self) -> Vec<ItemInventoryEntry> {
        let quantities: Collection<ItemInventoryEntry> =
            self.db.database("cdb").collection("item_quantities");

        let mut found = quantities
            .find(doc! { "item_name": self.item.name.clone()}, None)
            .await
            .unwrap();

        collect_results!(found)
    }

    pub async fn get_origins(&self) -> Vec<String> {
        let quantities: Collection<ItemInventoryEntry> =
            self.db.database("cdb").collection("item_quantities");

        let mut found = quantities
            .find(doc! { "item_name": self.item.name.clone()}, None)
            .await
            .unwrap();

        let mut hashset = HashSet::new();

        while let Some(res) = found.try_next().await.unwrap() {
            hashset.insert(res.origin);
        }

        hashset.into_iter().collect()
    }

    pub async fn get_variants(&self) -> Vec<String> {
        let variants: Collection<ItemVariantEntry> =
            self.db.database("cdb").collection("item_variants");
        let mut found_variants = variants
            .find(
                doc! {
                    "item_name": self.item.name.clone()
                },
                None,
            )
            .await
            .unwrap();

        let mut ret = vec!["Generic".to_string()];

        while let Some(doc) = found_variants.try_next().await.unwrap() {
            ret.push(doc.variant_name);
        }

        ret
    }

    pub async fn variant(&self, variant: &str) -> Option<Variant> {
        let variants: Collection<mongodb::bson::Document> =
            self.db.database("cdb").collection("variants");
        let res = variants
            .find_one(
                doc! {
                    "item": self.item.name.clone(),
                    "variant": variant
                },
                None,
            )
            .await
            .unwrap()?;

        Some(Variant {
            item: self.item.name.clone(),
            variant: variant.to_string(),
        })
    }

    pub async fn add_variant(&self, name: &str, amount: usize) {
        let variants: Collection<mongodb::bson::Document> =
            self.db.database("cdb").collection("variants");
        variants
            .insert_one(
                mongodb::bson::doc! {
                    "_id": format!("{}-{}", self.item.name.clone(), name),
                    "item": self.item.name.clone(),
                    "variant": name,
                    "amount": amount as i64
                },
                None,
            )
            .await
            .unwrap();
    }
}

#[derive(Debug, Clone)]
pub struct Price {
    pub value: f64,
    pub currency: String,
}

impl Price {
    pub fn new(value: f64, currency: &str) -> Self {
        Self {
            value,
            currency: currency.to_string(),
        }
    }

    fn parse(price: &str) -> Option<Self> {
        let (value, currency) = price.split_once(' ')?;

        Some(Self {
            value: value.parse().ok()?,
            currency: currency.to_string(),
        })
    }

    fn as_bson(&self) -> mongodb::bson::Bson {
        mongodb::bson::bson!({
            "value": self.value,
            "currency": self.currency.clone()
        })
    }
}

impl From<mongodb::bson::Document> for Price {
    fn from(value: mongodb::bson::Document) -> Self {
        Self {
            value: value.get_f64("value").unwrap(),
            currency: value.get_str("currency").unwrap().to_string(),
        }
    }
}

impl TryFrom<String> for Price {
    type Error = ();

    fn try_from(value: String) -> Result<Self, Self::Error> {
        Self::parse(&value).ok_or(())
    }
}

/// Represents a specific instance of an item with potential variations.
///
/// This struct is used to describe a particular variation or instance of an item
/// in the real world. It may include attributes or properties that deviate from
/// the standard definition of the item. For example, different colors, sizes, or
/// configurations.
pub struct Variant {
    pub item: String,
    pub variant: String,
}

pub struct BatchTransaction {
    pub uuid: String,
    pub transactions: Vec<String>,
}

impl BatchTransaction {
    pub fn new(transactions: Vec<String>) -> Self {
        Self {
            uuid: uuid::Uuid::new_v4().to_string(),
            transactions,
        }
    }

    fn as_doc(&self) -> mongodb::bson::Document {
        mongodb::bson::doc! {
            "_id": &self.uuid,
            "kind": "batch",
            "transactions": &self.transactions
        }
    }

    fn from(b: mongodb::bson::Document) -> Self {
        let uuid = b.get_str("_id").unwrap().to_string();
        let transactions = b
            .get_array("transactions")
            .unwrap()
            .into_iter()
            .map(|x| x.as_str().unwrap().to_string())
            .collect();
        Self { uuid, transactions }
    }
}

pub enum TransactionType {
    Batch(BatchTransaction),
    Normal(Transaction),
}

impl TransactionType {
    pub fn from(b: mongodb::bson::Document) -> Self {
        if let Ok(kind) = b.get_str("kind") {
            if kind == "batch" {
                return Self::Batch(BatchTransaction::from(b));
            }
        }
        Self::Normal(Transaction::from(b))
    }
}

pub struct Transaction {
    pub uuid: String,
    pub item: String,
    pub variant: String,
    pub price: Price,
    pub origin: String,
    pub timestamp: i64,
}

impl Transaction {
    pub fn new(item: &str, variant: &str, price: Price, origin: &str) -> Self {
        Self {
            uuid: uuid::Uuid::new_v4().to_string(),
            item: item.to_string(),
            variant: variant.to_string(),
            price,
            origin: origin.to_string(),
            timestamp: chrono::Utc::now().timestamp(),
        }
    }

    fn as_doc(&self) -> mongodb::bson::Document {
        mongodb::bson::doc! {
            "_id": &self.uuid,
            "item": &self.item,
            "variant": &self.variant,
            "price": self.price.as_bson(),
            "origin": &self.origin,
            "timestamp": self.timestamp
        }
    }

    fn from(b: mongodb::bson::Document) -> Self {
        let uuid = b.get_str("oid").unwrap().to_string();
        let item = b.get_str("item").unwrap().to_string();
        let variant = b.get_str("variant").unwrap().to_string();
        let origin = b.get_str("origin").unwrap().to_string();
        let timestamp = b.get_i64("timestamp").unwrap();
        let price = Price::from(b);
        Self {
            uuid,
            item,
            variant,
            price,
            origin,
            timestamp,
        }
    }
}

impl Variant {
    pub async fn demand(&self, uuid: &str) -> String {
        // todo : transaction_out + receipt
        todo!()
    }

    /// Records a supply transaction in the database.
    ///
    /// # Arguments
    ///
    /// * `amount` - The quantity of items supplied.
    /// * `price` - The price of the supplied items.
    /// * `origin` - The origin or source of the supplied items.
    ///
    /// # Returns
    ///
    /// Returns a UUID string representing the transaction.
    pub async fn supply(&self, amount: usize, price: Price, origin: &str) -> String {
        let db = get_mongo!();

        let mut ses = db.start_session(None).await.unwrap();

        let col: mongodb::Collection<mongodb::bson::Document> =
            ses.client().database("cdb").collection("transactions");

        let mut transactions = vec![];
        for _ in 0..amount {
            transactions.push(Transaction::new(
                &self.item,
                &self.variant,
                price.clone(),
                origin,
            ));
        }

        for transaction in &transactions {
            let r = col
                .insert_one_with_session(transaction.as_doc(), None, &mut ses)
                .await
                .unwrap();
        }

        // batch transaction
        let ret_uuid = if amount == 1 {
            transactions.first().unwrap().uuid.clone()
        } else {
            let batch =
                BatchTransaction::new(transactions.iter().map(|x| x.uuid.clone()).collect());
            let col: mongodb::Collection<mongodb::bson::Document> = ses
                .client()
                .database("cdb")
                .collection("transactions_batch");
            col.insert_one_with_session(batch.as_doc(), None, &mut ses)
                .await
                .unwrap();
            batch.uuid
        };

        // todo : transaction overlap cache -> scale

        ses.commit_transaction().await.unwrap();

        ret_uuid
    }
}

struct InventoryCache {}

impl InventoryCache {
    pub async fn push(uuid: &str, ses: &mut ClientSession) {
        // todo : if not exists?
        let update = doc! { "$push": { "transactions": uuid } };

        let options = mongodb::options::FindOneAndUpdateOptions::builder()
            .upsert(true)
            .return_document(mongodb::options::ReturnDocument::After)
            .build();

        let result = ses
            .client()
            .database("cdb")
            .collection::<mongodb::bson::Document>("cache")
            .find_one_and_update(doc! { "_id": "inventory"}, update, options)
            .await
            .unwrap();
    }

    pub async fn remove(uuid: &str) {
        // todo : remove from cache
    }
}

pub struct ItemDB {
    index: mdq::Index,
    mongodb: mongodb::Client,
}

impl ItemDB {
    pub async fn new(dir: &str, mongodb: &str) -> Self {
        // scan for markdown item entries
        let index = mdq::Index::new(dir, true);
        let mongodb = get_mongo!();

        for item in &index.documents {
            let item = ItemEntry::new(item.clone());
            item.init_db(&mongodb).await;
        }

        Self { index, mongodb }
    }

    /// Retrieves an item by name
    pub fn get_item(&self, item: &str) -> Option<Item> {
        Some(Item::new(
            self.mongodb.clone(),
            self.index
                .documents
                .iter()
                .map(|x| ItemEntry::new(x.clone()))
                .find(|x| x.name == item)?,
        ))
    }

    /// Get all items
    pub fn items(&self) -> Vec<String> {
        let mut ret = vec![];
        for item in &self.index.documents {
            let item = ItemEntry::new(item.clone());
            ret.push(item.name);
        }
        ret
    }
}
